/***********************************/ 
/* oroborus (c) Ken Lynch Jan 2001 */ 
/* Distributed under the GPL       */ 
/* See LICENSE for more details    */ 
/***********************************/ 
 
#include "oroborus.h" 
#define ICONSTEPS 7
 
/* 
 * 
 * Handle different window gravities 
 * 
 */ 

void gravitate_client(Client *c, int grav) 
{ 
  int dx=0, dy=0; 
 
#ifdef DEBUG 
  printf("gravitate_client\n"); 
#endif 
 
  switch(GRAVITY) 
  { 
  case NorthWestGravity: 
    dx=LW; 
    dy=TH; 
    break; 
  case NorthGravity: 
    dx=(LW+RW)/2; 
    dy=TH; 
    break; 
  case NorthEastGravity: 
    dx=-RW; 
    dy=TH; 
    break; 
  case WestGravity: 
    dx = LW; 
    dy=(TH+BH)/2; 
    break; 
  case EastGravity: 
    dx=-RW; 
    dy=(TH+BH)/2; 
    break;  
  case SouthWestGravity: 
    dx=LW; 
    dy=-BH; 
    break; 
  case SouthGravity: 
    dx=(LW+RW)/2; 
    dy=-BH; 
    break; 
  case SouthEastGravity: 
    dx=-RW; 
    dy=-BH; 
    break; 
  case CenterGravity: 
    dx=(LW+RW)/2; 
    dy=(TH+BH)/2; 
    break; 
  } 
  c->x=c->x+(dx*grav); 
  c->y=c->y+(dy*grav); 
} 
 
/* 
 * 
 * Create a new frame part 
 * 
 */ 

Window create_frame_part(Window win, int x, int y, int w, int h, int mapit) 
{ 
  Window new;   
 
#ifdef DEBUG 
  printf("create_frame_part %#lx (%#lx)\n",win,lastfocus); 
#endif 
 
  new=XCreateSimpleWindow(dpy, win, x, y, w, h, 0, 0, 0);

  XSelectInput(dpy, new, ButtonPressMask|ButtonReleaseMask); 
  if( mapit ) XMapWindow(dpy, new); 
  return new; 
} 
 
/* 
 * 
 * Create a new client frame 
 * 
 */ 

void create_new_client(Window w, int reposition) 
{ 
  Client *c=NULL; 
  XWindowAttributes attr; 
  XSetWindowAttributes wattr; 
  PropMwmHints *mhints=NULL; 
  XWMHints *hints=NULL; 
  long dummy; 
  XWindowChanges wc; 
  int x, y, xmax, ymax, xmin, ymin, head=0; 
 
#ifdef DEBUG 
  printf("create_new_client %#lx (%#lx)\n",w,lastfocus); 
#endif 
 
  if((c=malloc(sizeof *c))==NULL)
  {
    fprintf(stderr, "%s: Unable to allocate memory for new client.\n", PROGNAME);
    exit(1);
  }
 
  /* Initialize client values */ 
  c->window=w; 
  c->ignore_unmap=0; 
  c->hide_state=BUTTON_UNPRESSED; 
  c->icon_state=BUTTON_UNPRESSED;
  c->close_state=BUTTON_UNPRESSED; 
  c->can_max=True; 
  c->has_frame=True; 
  c->has_focus=False; 
  c->is_icon=0;
  c->name_x=-1;
  c->name_y=-1;
  c->name_w=-1;
  c->name_h=-1;
  c->icon_x=-1000;
  c->icon_y=-1000;
  c->icon_h=-1;
  c->icon_w=-1;
  c->icontxt_x=-1;
  c->icontxt_y=-1;
  c->icontxt_w=-1;
  c->icontxt_h=-1;

#ifdef I18N_SUPPORT 
  if(use_fontset)
    /* call wrapper function for I18N */ 
    XFetchNameWrapper(dpy, c->window, &c->name); 
  else
    XFetchName(dpy, c->window, &c->name);
#else 
  XFetchName(dpy, c->window, &c->name); 
#endif /* I18N */ 

  c->size=XAllocSizeHints(); 
  XGetWMNormalHints(dpy, w, c->size, &dummy); 
 
  /* Get size and position of client */ 
  XGetWindowAttributes(dpy, c->window, &attr); 
   
  /* Handle clients with border set > 0 and adjust position accordingly */ 
  if(attr.border_width>0) 
  { 
    switch(GRAVITY) 
    { 
    case SouthEastGravity: 
      attr.x=attr.x+attr.border_width*2; 
      attr.y=attr.y+attr.border_width*2; 
      break; 
    case SouthWestGravity: 
    case SouthGravity: 
      attr.y=attr.y+attr.border_width*2; 
      break; 
    case NorthEastGravity: 
    case EastGravity: 
      attr.x=attr.x+attr.border_width*2; 
      break; 
    } 
  } 
  c->x=attr.x; 
  c->y=attr.y; 
  c->width=attr.width; 
  c->height=attr.height; 
 
  /* Set initial client state */ 
  if(attr.map_state!=IsViewable) 
  { 
    /* Default state NormalState */ 
    set_wm_state(c->window, NormalState); 
    hints=XGetWMHints(dpy, c->window); 
    if(hints) 
    { 
      /* Sets the initial window state if set by client */ 
      if(hints->flags&StateHint) 
        set_wm_state(c->window, hints->initial_state); 
 
      XFree(hints); 
    } 
  } 
 
  /* Get MWM hints */ 
  mhints=get_mwm_hints(c->window); 
  if(mhints) 
  { 
#ifdef DEBUG
    printf("decorations: %#lx\n",mhints->decorations);
    printf("DECOR_BORDER, DECOR_TITLE: %#lx , %#lx\n",MWM_DECOR_BORDER,MWM_DECOR_TITLE); 
#endif
    if((mhints->flags&MWM_HINTS_DECORATIONS)!=0 && (mhints->decorations&MWM_DECOR_ALL)==0) 
    { 
      /* Window has no border if *neither* MWM_DECOR_BORDER nor MWM_DECOR_TITLE are set */ 
      c->has_frame=((mhints->decorations&(MWM_DECOR_BORDER|MWM_DECOR_TITLE))!=0); 

      /* Set if client can be maximized or not */ 
      c->can_max=((mhints->decorations&MWM_DECOR_MAXIMIZE)!=0); 
    } 
    XFree(mhints); 
  } 
 
  /* Set event mask for client */ 
  XSelectInput(dpy, c->window, PropertyChangeMask|FocusChangeMask); 
 
#ifdef SHAPE_SUPPORT 
  /* Select shape events */ 
  if(shape) 
    XShapeSelectInput(dpy, c->window, ShapeNotifyMask); 
#endif /* SHAPE */ 
 
  /* Create frame window */ 
  wattr.override_redirect=True; 
  wattr.event_mask=SubstructureNotifyMask|SubstructureRedirectMask|ButtonPressMask|ButtonReleaseMask|EnterWindowMask|LeaveWindowMask; 
  c->frame=XCreateWindow(dpy, root, FX, FY, FW, FH, 0, DefaultDepth(dpy, screen), CopyFromParent, DefaultVisual(dpy, screen), CWOverrideRedirect|CWEventMask, &wattr); 
  XDefineCursor(dpy, c->frame, normal_curs); 

  /* This ignores the unmap event when reparenting */ 
  if(attr.map_state==IsViewable) 
    c->ignore_unmap++; 
 
  /* Reparent client window */ 
  XReparentWindow(dpy, c->window, c->frame, LW, TH); 
    
  /* Create frame parts if has border and/or title */ 
    /* Top frame parts */ 
    c->title=create_frame_part(c->frame, TX, TY, TW<1?1:TW, TH, c->has_frame); 
    c->top_left=create_frame_part(c->frame, TLX, TLY, TLW, TLH, c->has_frame); 
    c->top_right=create_frame_part(c->frame, TRX, TRY, TRW, TRH, c->has_frame);  
   
  if(c->has_frame) 
  { 
    /* Side frame parts */ 
    c->left=create_frame_part(c->frame, LX, LY, LW, LH, 1); 
    c->right=create_frame_part(c->frame, RX, RY, RW, RH, 1); 
     
    /* Bottom frame parts */ 
    c->bottom=create_frame_part(c->frame, BX, BY, BW, BH, 1); 
    c->bottom_left=create_frame_part(c->frame, BLX, BLY, BLW, BLH, 1); 
    c->resize=create_frame_part(c->frame, RSX, RSY, RSW, RSH, 1); 
   
    /* Top frame parts 
    c->title=create_frame_part(c->frame, TX, TY, TW<1?1:TW, TH); 
    c->top_left=create_frame_part(c->frame, TLX, TLY, TLW, TLH); 
    c->top_right=create_frame_part(c->frame, TRX, TRY, TRW, TRH);   */
   
    /* Buttons */ 
    c->hide=create_frame_part(c->frame, HX, HY, HW, HH, 1); 
    c->icon=create_frame_part(c->frame, IX, IY, IW, IH, 1);   
    c->close=create_frame_part(c->frame, CX, CY, CW, CH, 1);
  } 
 
  /* Move the client to position set by hints */ 
  if((c->size->flags & (PPosition|USPosition)) || !reposition) 
  {
#ifdef XINERAMA_SUPPORT
    if(xinerama)
    {
      /* If window positioned within 5 pixels of center of display, consider it to be centered on the current head */
      if(abs((c->x+(c->width/2))-(DisplayWidth(dpy, screen)/2))<5 && abs((c->y+(c->height/2))-(DisplayHeight(dpy, screen)/2))<5)
      {
        get_mouse_xy(&x, &y);
        head=find_head(x, y);
        c->x=screen_x(head)+(screen_width(head)/2)-(c->width/2);
        c->y=screen_y(head)+(screen_height(head)/2)-(c->height/2);
      }
      head=find_head(c->x, c->y);
    }
#endif /* XINERAMA */
 
    xmin=screen_x(head)+LW; 
    ymin=screen_y(head)+TH; 
    xmax=(screen_x(head)+screen_width(head))-c->width-RW; 
    ymax=(screen_y(head)+screen_height(head))-c->height-BH; 
   
    gravitate_client(c, APPLY_GRAVITY); 
    wc.x=c->x; 
    wc.y=c->y; 
 
    /* Keep window fully inside desktop */ 
    if(wc.x>xmax) wc.x=xmax; 
    if(wc.y>ymax) wc.y=ymax; 
    if(wc.x<xmin) wc.x=xmin; 
    if(wc.y<ymin) wc.y=ymin; 
  } 
 
  /* Otherwise follow window placement policy */ 
  else 
  { 
#ifdef XINERAMA_SUPPORT
    if(xinerama)
    {
      get_mouse_xy(&x, &y);
      head=find_head(x, y);
    }
#endif /* XINERAMA */

    xmin=screen_x(head)+LW+config.left_margin; 
    ymin=screen_y(head)+TH+config.top_margin; 
    xmax=(screen_x(head)+screen_width(head))-c->width-RW-config.right_margin; 
    ymax=(screen_y(head)+screen_height(head))-c->height-BH-config.bottom_margin; 
    if(config.window_placement==CENTER_ROOT) 
    { 
      x=screen_x(head)+(screen_width(head)/2); 
      y=screen_y(head)+(screen_height(head)/2); 
    } 
    else 
      get_mouse_xy(&x, &y); 
 
    wc.x=x-(c->width/2); 
    wc.y=y-(c->height/2); 
 
    /* Keep window fully inside desktop */ 
    if(wc.x>xmax) wc.x=xmax; 
    if(wc.y>ymax) wc.y=ymax; 
    if(wc.x<xmin) wc.x=xmin; 
    if(wc.y<ymin) wc.y=ymin; 
  } 
  wc.height=c->height; 
  wc.width=c->width; 
  wc.border_width=0; 
  configure_client(c, CWHeight|CWWidth|CWX|CWY|CWBorderWidth, &wc, False); 
  redraw_client_frame(c); 
 
  /* Set client to correct state */   
  if(attr.map_state==IsViewable) 
  { 
    if(get_wm_state(c->window)==IconicState) {
      raise_client(c);
      hide_client(c);
    } else {
      raise_client(c); }
  } 
  else 
  { 
    if(get_wm_state(c->window)==NormalState) 
      raise_client(c); 
  } 
 
  /* Button grab for click-to-focus */ 
  XGrabButton(dpy, AnyButton, AnyModifier, c->frame, False, ButtonPressMask|ButtonReleaseMask, GrabModeSync, GrabModeAsync, None, None); 
 
#ifdef KEYBOARD_SUPPORT 
  grab_keys(c->window); 
#endif /* KEYBOARD */ 
 
#ifdef WORKSPACE_SUPPORT 
  set_gnome_hint(c->window, WIN_WORKSPACE, workspace); 
#endif /* WORKSPACE */ 
 
  /* Add client to client list */ 
  if(client_list==NULL) 
  { 
    client_list=c; 
    c->prev=c; 
    c->next=c; 
  } 
  else 
  { 
    c->prev=client_list->prev; 
    c->next=client_list; 
    client_list->prev->next=c; 
    client_list->prev=c; 
  } 
  client_count++; 
  update_gnome_client_list(); 
 
  /* Check if we want to focus new clients */ 
  if((c->is_icon==0) && config.focus_new && !(get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS)) 
    set_focus_client(c); 

#ifdef DEBUG
  printf("  exit create_new_client\n");
#endif
} 
 
/*  
 *  
 * Get client frame from window ID 
 * 
 */ 

Client *get_client_from_window(Window w, int win) 
{ 
  Client *c=NULL; 
  Client *ct=NULL;
  Client *cf=NULL;
  Client *cw=NULL;
  long i; 
 
#ifdef DEBUG 
  printf("get_client_from_window %#lx %d %d\n",w,win,client_count); 
#endif 

  if(w==0 || client_count==0) return NULL;
 
  c=client_list; 
  for(i=0;i<client_count;i++) 
  { 
#ifdef DEBUG
   printf("   checking client [%#lx] %#lx _%#lx_\n",c->frame,c->window,c->title);
#endif
    if(win==TITLE) 
    { 
      if(c->title==w) {
        ct=c;
        if(w!=lastenter)
          return c; 
      }
    } 
    else if(win==FRAME) 
    { 
      if(c->frame==w) {
        cf=c;
        if(w!=lastenter)
          return c;  
      }
    }  
    else 
    { 
      if(c->window==w) { 
        cw=c;
        if(w!=lastenter)
          return c; 
      }
    } 
    c=c->next; 
  } 

  if(w==lastenter) {
    if(win==TITLE) {
      if(ct==NULL) {
#ifdef DEBUG
         printf("  failed to find title for lastenter - setting to zero\n");
#endif
         lastenter=0; }
      return ct;
    } else if(win==FRAME) {
      if(cf==NULL) {
#ifdef DEBUG
         printf("  failed to find frame for lastenter - setting to zero\n");
#endif
         lastenter=0; }
      return cf;
    } else {
      if(cw==NULL) {
#ifdef DEBUG
         printf("  failed to find window for lastenter - setting to zero\n");
#endif
         lastenter=0; }
    }
  }

  return cw; 
} 
 
/* 
 * 
 * Get client with focus 
 * 
 */ 

Client *get_focus_client() 
{ 
  Client *c; 
  Window focus_win; 
  int dummy; 
 
#ifdef DEBUG 
  printf("get_focus_client (%#lx)\n",lastfocus); 
#endif 
 
  XGetInputFocus(dpy, &focus_win, &dummy); 
  c=get_client_from_window(focus_win, WINDOW); 

#ifdef DEBUG
  if(c==NULL)
    printf("  got no client for focus_win %#lx (%#lx)\n",focus_win,lastfocus);
  else
    printf("  got focus_win: %#lx (%#lx) [%#lx] %#lx\n",focus_win,lastfocus,c->frame,c->window);
#endif

  hacking_focus=False;
  if(c!=NULL)
    lastfocus=focus_win;
  else if(focus_win==check_win)
    lastfocus=0;
  else {
    c=get_client_from_window(lastfocus, WINDOW);
    if(c!=NULL)
      hacking_focus=True;
    else
      lastfocus=0;
  }

  return c;   
} 
 
/* 
 * 
 * Get next client in client list 
 * 
 */ 

Client *get_next_client(Client *c) 
{ 
  long i; 
 
#ifdef DEBUG 
  printf("get_next_client\n"); 
#endif 
 
  if(c!=NULL) 
  { 
    for(i=0;i<client_count;i++) 
    { 
      c=c->next; 
#ifdef WORKSPACE_SUPPORT 
      if(!get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS && get_gnome_hint(c->window, WIN_WORKSPACE)==workspace) 
#else 
      if(!get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS) 
#endif /* WORKSPACE */ 
        return c; 
    } 
  } 
  return NULL; 
} 

/* 
 * 
 * Get previous client in client list 
 * 
 */ 

Client *get_prev_client(Client *c) 
{ 
  long i; 
 
#ifdef DEBUG 
  printf("get_prev_client\n"); 
#endif 
 
  if(c!=NULL) 
  { 
    for(i=0;i<client_count;i++) 
    { 
      c=c->prev; 
#ifdef WORKSPACE_SUPPORT 
      if(!get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS && get_gnome_hint(c->window, WIN_WORKSPACE)==workspace) 
#else 
      if(!get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS) 
#endif /* WORKSPACE */ 
        return c; 
    } 
  } 
  return NULL;
}
 
/* 
 * 
 * Get top-most client 
 * 
 */ 

Client *get_top_most_client(int layer, int get_icons) 
{ 
  Window *windows, w1, w2; 
  int count, i; 
  Client *top=NULL; 
#ifdef LAYER_SUPPORT 
  Client *c=NULL; 
#endif /* LAYER */ 
  XWindowAttributes attr; 
 
#ifdef DEBUG 
  printf("get_top_most_client\n"); 
#endif 
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
#ifdef LAYER_SUPPORT 
  for(i=0;i<count;i++) 
  { 
    c=get_client_from_window(windows[i], FRAME); 
    if(c!=NULL && (c->is_icon==0 || get_icons)) 
    { 
      XGetWindowAttributes(dpy, c->window, &attr); 
      if(attr.map_state==IsViewable || get_icons) 
      { 
        if(get_gnome_hint(c->window, WIN_LAYER)<=layer) 
          top=c; 
        else 
          break; 
      } 
    } 
  } 
#else 
  for(i=count;i<=0;i--) 
  { 
    c=get_client_from_window(windows[i], FRAME); 
    if(c!=NULL && (c->is_icon==0 || get_icons)) 
    { 
      XGetWindowAttributes(dpy, c->window, &attr); 
      if(attr.map_state==IsViewable || get_icons) { 
        top=c;
        break; }
    } 
  } 
#endif /* LAYER */ 
  if(windows) XFree(windows); 
  return top; 
} 
 
/* 
 * 
 * Get bottom-most client 
 * 
 */ 

Client *get_bottom_most_client(int layer, int get_icons) 
{ 
  Window *windows=NULL, w1, w2; 
  int count, i; 
  Client *c=NULL; 
  Client *bot=NULL;
  XWindowAttributes attr; 
 
#ifdef DEBUG 
  printf("get_bottom_most_client\n"); 
#endif 
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
  for(i=0;i<count;i++) 
  { 
    c=get_client_from_window(windows[i], FRAME); 
    if(c!=NULL && (c->is_icon==0 || get_icons)) 
    { 
      XGetWindowAttributes(dpy, c->window, &attr); 
      if(attr.map_state==IsViewable || get_icons) 
      { 
#ifdef LAYER_SUPPORT 
        if(get_gnome_hint(c->window, WIN_LAYER)>=layer) { 
          bot=c;
          break; }
#else 
        break; 
#endif /* LAYER */ 
      } 
    } 
  } 
  if(windows) XFree(windows);  
  return bot; 
} 
 
/* 
 *  
 * Raise client window 
 * 
 */ 

void raise_client(Client *c) 
{ 
  XWindowChanges wc; 
 
#ifdef DEBUG 
  printf("raise_client (%#lx)\n",lastfocus); 
#endif 
 
  if(c==NULL) 
  { 
#ifdef DEBUG 
    printf("  client is NULL\n"); 
#endif 
  } 
  else 
  { 
#ifdef DEBUG 
    printf("  raising client [%#lx] %#lx\n",c->frame,c->window); 
#endif 
 
    if(c->is_icon==0)
      show_client(c); 
    wc.stack_mode=Above; 
    configure_client(c, CWStackMode, &wc, c->is_icon!=0); 
  } 
} 
 
/*
 *
 * Entered/Left a client
 *
 */

void entered_client(Client *c, int can_focus)
{
#ifdef DEBUG
  printf("entered_client [%#lx] %#lx (%#lx)\n",c==NULL?0:c->frame,c==NULL?0:c->window,lastfocus);
#endif

  if(overicon!=NULL) {
     if(overicon->is_icon!=2) {
       overicon=NULL;
     } else if(overicon!=c) {
       overicon->is_icon=1;
       if( overicon->icon_w!=100 )
          resize_client(overicon,100,overicon->icon_h,True);
       redraw_client_frame(overicon);
       lower_client(overicon);
       overicon=NULL; }
  }

  if(c==NULL)
  {
    if(config.focus_style==FOCUS_SLOPPY && can_focus)
      set_focus_client(NULL);
  }
  else if(c->is_icon!=0)
  {
    c->is_icon=2;
    if( c->icontxt_w>=TW-6 )
       resize_client(c,c->icontxt_w+6-LW-RW+TLW+TRW,c->icon_h,True);
    redraw_client_frame(c);
    raise_client(c);
    overicon=c;
  }
  else if(config.focus_style==FOCUS_SLOPPY && can_focus)
  {
    set_focus_client(c);
  }

}

/* 
 *  
 * Give client window keyboard focus 
 * 
 */ 

void set_focus_client(Client *c) 
{ 
#ifdef DEBUG 
  printf("set_focus_client (%#lx)\n",lastfocus); 
#endif 
 
  if(c==NULL) 
  { 
#ifdef DEBUG 
    printf("  client is NULL\n"); 
#endif 

    XSetInputFocus(dpy, check_win, RevertToNone, CurrentTime); 
    lastfocus=0;
  } 
  else if(c->is_icon==0)
  { 
#ifdef DEBUG 
    printf("  focusing client [%#lx] %#lx\n",c->frame,c->window); 
#endif 

    XSetInputFocus(dpy, c->window, RevertToNone, CurrentTime); 
    if(lastfocus!=c->window) {
       lastfocus=c->window;
       if(config.focus_style==FOCUS_CLICK)
         raise_client(c);
       redraw_client_frame(c);
    }
  } 
} 
 
/* 
 * 
 * Lower client window 
 * 
 */ 

void lower_client(Client *c) 
{ 
  XWindowChanges wc; 
 
#ifdef DEBUG 
  printf("lower_client (%#lx)\n",lastfocus); 
#endif 
 
  if(c==NULL) {
#ifdef DEBUG
    printf("  client is NULL\n");
#endif
    return;
  }

#ifdef DEBUG
  printf("  lowering client [%#lx] %#lx\n",c->frame,c->window);
#endif

  wc.stack_mode=Below; 
  configure_client(c, CWStackMode, &wc, c->is_icon!=0); 
} 
 
/* 
 *  
 * Show client window 
 * 
 */ 
void show_client(Client *c) 
{ 
#ifdef DEBUG 
  printf("show_client (%#lx)\n",lastfocus); 
#endif 
 
  if(c==NULL) {
#ifdef DEBUG
    printf("  client is NULL\n");
#endif
    return;
  } 

#ifdef DEBUG
  printf("  showing client [%#lx] %#lx\n",c->frame,c->window);
#endif
  c->is_icon=0;
  XMapWindow(dpy, c->frame); 
  XMapWindow(dpy, c->window); 
  set_wm_state(c->window, NormalState); 
} 
 
/* 
 * 
 * Hide client window 
 * 
 */ 

void old_hide_client(Client *c) 
{ 
#ifdef DEBUG 
  printf("hide_client (old)\n"); 
#endif 

  if(c==NULL) {
#ifdef DEBUG
    printf("  client is NULL\n");
#endif
    return;
  }

#ifdef DEBUG 
  printf("  hiding client [%#lx] %#lx\n",c->frame,c->window);
#endif

  /* We need to ignore one of the unmaps, otherwise our client gets deleted */ 
  c->ignore_unmap++; 
  XUnmapWindow(dpy, c->frame); 
  XUnmapWindow(dpy, c->window); 
  set_wm_state(c->window, IconicState); 
} 
 
/*  
 * 
 * Configure client - the main resizing, raising, lowering function 
 * 
 */ 

void configure_client(Client *c, unsigned int mask, XWindowChanges *wc, int frame_only) 
{ 
  XConfigureEvent ce; 
  int w, w1, h, h1, orig_is_icon;
 
#ifdef DEBUG 
  printf("configure_client (%#lx)\n",lastfocus); 
#endif 

  if(c==NULL) {
#ifdef DEBUG
  printf("  client is NULL\n");
#endif
    return;
  }

#ifdef DEBUG
  printf("  configuring client [%#lx] %#lx\n",c->frame,c->window);
#endif

  orig_is_icon=c->is_icon;
  if(!frame_only)
    c->is_icon=0;

  if(c->is_icon!=0) {
    h=c->icon_h;
    w=c->icon_w;
    if(mask&CWX) c->icon_x=wc->x;
    if(mask&CWY) c->icon_y=wc->y;
  } else {
    h=c->height;
    w=c->width;
    if(mask&CWX) c->x=wc->x;
    if(mask&CWY) c->y=wc->y;
  }

  if(mask&CWWidth) 
  { 
    w=wc->width; 
 
    /*printf("Setting width: %d , %d , %d , %d , %d\n",w,c->icontxt_x,c->name_w,c->width,c->icon_w);*/

    if( c->is_icon!=0 && frame_only) {
      c->icon_w=w;
    } else {

    /* Get resize increments */ 
    if(c->size->flags & PResizeInc) 
    { 
      w1=(w-c->size->min_width)/c->size->width_inc; 
      w=c->size->min_width+(w1*c->size->width_inc); 
    } 
    /* Get min size */ 
    if(c->size->flags & PMinSize) 
      if(w<c->size->min_width) w=c->size->min_width; 
    /* Get max size */ 
    if(c->size->flags & PMaxSize) 
      if(w>c->size->max_width) w=c->size->max_width; 
 
    /* Make sure we never set width to <1 */ 
    if(w<1) w=1; 

      c->width=w; }
  } 

  if(mask&CWHeight) 
  { 
    h=wc->height; 
 
    if( c->is_icon!=0 && frame_only ) {
      c->icon_h=h;
    } else {

    /* Get resize increments */ 
    if(c->size->flags & PResizeInc) 
    { 
      h1=(h-c->size->min_height)/c->size->height_inc; 
      h=c->size->min_height+(h1*c->size->height_inc); 
    } 
    /* Get min size */ 
    if(c->size->flags & PMinSize) 
      if(h<c->size->min_height) h=c->size->min_height; 
    /* Get max size */ 
    if(c->size->flags & PMaxSize) 
      if(h>c->size->max_height) h=c->size->max_height; 
 
    /* Make sure we never set height to <1 */ 
    if(h<1) h=1; 

      c->height=h; }
  } 
 
#ifdef LAYER_SUPPORT 
  /* Deal with window layers */ 
  if(mask&CWStackMode) 
  { 
    Client *sibling=NULL; 
 
    /* Takes care of layers */ 
    switch(wc->stack_mode) 
    { 
    case Above: 
      /* Window wants to be raised to top of its layer */ 
      sibling=get_top_most_client(get_gnome_hint(c->window, WIN_LAYER),True); 
      if(sibling==NULL) 
        wc->stack_mode=Below; 
      break; 
    case Below: 
      /* Window wants to be lowered to bottom of its layer */ 
      sibling=get_bottom_most_client(get_gnome_hint(c->window, WIN_LAYER),True); 
      break; 
    } 
    if(sibling==NULL) 
    { 
      if(mask&CWSibling) mask=mask^CWSibling; 
    } 
    else if(sibling==c) 
    { 
      if(mask&CWStackMode) mask=mask^CWStackMode; 
      if(mask&CWSibling) mask=mask^CWSibling; 
    } 
    else 
    { 
      mask=mask|CWSibling; 
      wc->sibling=sibling->frame; 
    } 
#ifdef WORKSPACE_SUPPORT 
    set_gnome_hint(c->window, WIN_WORKSPACE, workspace); 
#endif /* WORKSPACE */ 
  } 
#endif /* LAYER */ 
 
  wc->x=FX; 
  wc->y=FY; 
  wc->width=w+LW+RW;
  wc->height=FH; 
  XConfigureWindow(dpy, c->frame, mask, wc); 

  if(!frame_only)
  {
    wc->x=LW;
    wc->y=TH;
    wc->width=c->width;
    wc->height=c->height;

    /* We don't want any stacking order changes on our actual client */ 
    if(mask&CWStackMode) mask=mask^CWStackMode; 
    if(mask&CWSibling) mask=mask^CWSibling;

    XConfigureWindow(dpy, c->window, mask, wc);
  }

  /* If size has changed */ 
  if(mask&(CWHeight|CWWidth)) 
  { 
    if( c->has_frame || (c->is_icon!=0) ) 
    { 
      /* Reposition and resize frame parts */ 

      if( w+LW+RW-TRW-TLW>1 )
        XMoveResizeWindow(dpy, c->title, TX, TY, w+LW+RW-TRW-TLW, TH); 
      else
        XMoveResizeWindow(dpy, c->title, TX, TY, 1, TH);
      XMoveResizeWindow(dpy, c->top_left, TLX, TLY, TLW, TLH); 
      XMoveResizeWindow(dpy, c->top_right, w+LW+RW-TRW, TRY, TRW, TRH);

      if( c->is_icon!=0 ) {
        XMoveResizeWindow(dpy, c->close, -100, -100, CW, CH);
        XMoveResizeWindow(dpy, c->icon, -100, -100, IW, IH);
        XMoveResizeWindow(dpy, c->hide, -100, -100, HW, HH);
      } else if( frame_only) {
        XMoveResizeWindow(dpy, c->close, CX, CY, CW, CH);
        XMoveResizeWindow(dpy, c->icon, IX, IY, IW, IH);
        XMoveResizeWindow(dpy, c->hide, HX, HY, HW, HH);
      } else {
        XMoveResizeWindow(dpy, c->left, LX, LY, LW, LH);
        XMoveResizeWindow(dpy, c->right, RX, RY, RW, RH);
        XMoveResizeWindow(dpy, c->bottom_left, BLX, BLY, BLW, BLH);
        XMoveResizeWindow(dpy, c->bottom, BX, BY, BW, BH);
        XMoveResizeWindow(dpy, c->resize, RSX, RSY, RSW, RSH);
        XMoveResizeWindow(dpy, c->close, CX, CY, CW, CH);
        XMoveResizeWindow(dpy, c->icon, IX, IY, IW, IH);
        XMoveResizeWindow(dpy, c->hide, HX, HY, HW, HH);
      }
      redraw_client_frame(c);
    }

#ifdef SHAPE_SUPPORT 
    set_client_shape(c);
#endif /* SHAPE */ 
  }
 
  /* Notify client of window changes if changes to position or size */ 
  if(mask&(CWX|CWY|CWHeight|CWWidth) && !frame_only && (c->is_icon==0) ) 
  { 
    ce.type=ConfigureNotify; 
    ce.event=c->window; 
    ce.window=c->window; 
    ce.x=c->x; 
    ce.y=c->y; 
    ce.width=c->width; 
    ce.height=c->height; 
    ce.border_width=0; 
    ce.above=None; 
    ce.override_redirect=0; 
    XSendEvent(dpy, c->window, False, StructureNotifyMask, (XEvent *)&ce); 
  } 

  /*c->is_icon=orig_is_icon;*/
} 
 
/* 
 * 
 * Move client to new position 
 * 
 */ 

void move_client(Client *c, int x, int y) 
{ 
  XWindowChanges wc; 
 
#ifdef DEBUG 
  printf("move_client (%#lx)\n",lastfocus); 
  printf("  moving client [%#lx] %#lx\n",c->frame,c->window); 
#endif 
 
  wc.x=x; 
  wc.y=y; 
  configure_client(c, CWX|CWY, &wc, c->is_icon!=0); 
} 
 
/* 
 * 
 * Resize client 
 * 
 */ 

void resize_client(Client *c, int w, int h, int frame_only) 
{ 
  XWindowChanges wc; 
 
#ifdef DEBUG 
  printf("resize_client (%#lx)\n",lastfocus); 
  printf("  resizing client [%#lx] %#lx\n",c->frame,c->window); 
#endif 
 
  wc.width=w; 
  wc.height=h; 
  configure_client(c, CWWidth|CWHeight, &wc, frame_only || (c->is_icon!=0)); 
} 
 
/* 
 * 
 * Toggle client maximized 
 * 
 */ 

void toggle_client_maximized(Client *c) 
{ 
  XWindowChanges wc; 
 
  if( c->is_icon!=0 )
    return;

#ifdef DEBUG 
  printf("toggle_client_maximized %d\n",c->can_max); 
#endif 
 
  if(c->can_max) 
  { 
    if(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_MAXIMIZED) 
    { 
      c->x=c->old_x; 
      c->y=c->old_y; 
      c->width=c->old_width; 
      c->height=c->old_height; 
    } 
    else 
    { 
      int head=0; 
 
#ifdef XINERAMA_SUPPORT 
      if(xinerama)
        head=find_head(c->x+c->width/2, c->y+c->height/2); 
#endif /* XINERAMA */ 
 
      c->old_x=c->x; 
      c->old_y=c->y; 
      c->old_width=c->width; 
      c->old_height=c->height; 
      c->x=screen_x(head)+LW+config.left_margin; 
      c->y=screen_y(head)+TH+config.top_margin; 
      c->width=screen_width(head)-LW-RW-config.left_margin-config.right_margin; 
      c->height=screen_height(head)-TH-BH-config.top_margin-config.bottom_margin;
    } 
    wc.x=c->x; 
    wc.y=c->y; 
    wc.height=c->height; 
    wc.width=c->width; 
    configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, c->is_icon!=0); 
    set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_MAXIMIZED); 
  } 
} 
 
/* 
 *  
 * Toggle client shaded 
 * 
 */ 

void toggle_client_shaded(Client *c) 
{ 

  if(c->is_icon!=0)
    return;

#ifdef DEBUG 
  printf("toggle_client_shaded\n"); 
#endif 
 
  if(c->has_frame) 
  { 
    set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_SHADED); 

    resize_client(c, c->width, c->height, True);
    redraw_client_frame(c); 
  }

}


void hide_client(Client *c) 
{ 
  XWindowChanges wc;
  int i, nowt, nowx, nowy, noww, endt, endx, endy, endw;
  int shaded=0;
  Client *newc=NULL;

#ifdef DEBUG 
  printf("hide_client (mine)\n"); 
#endif 
 
    newc=get_focus_client();

    if( c->is_icon!=0 ) {
       if(newc) {
         set_focus_client(NULL);
         redraw_client_frame(newc);
         newc=NULL; }
       nowt=c->icontxt_x;
       nowx=c->icon_x;
       nowy=c->icon_y;
       noww=c->icon_w;
       endt=c->name_x;
       endx=c->x;
       endy=c->y;
       endw=c->width;
       c->is_icon=2;
       for(i=ICONSTEPS-1;i>=0;i--) {
         wc.x=endx-(endx-nowx)*i/ICONSTEPS;
         wc.y=endy-(endy-nowy)*i/ICONSTEPS;
         wc.width=endw-(endw-noww)*i/ICONSTEPS;
         wc.height=c->icon_h;
         c->icontxt_x=endt-(endt-nowt)*i/ICONSTEPS;
         configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, True); 
         /*usleep(5000000);*/
       }
       wc.x=c->x;
       wc.y=c->y;
       wc.width=c->width;
       wc.height=c->height;
       c->is_icon=0;
       c->icon_x=nowx;
       c->icon_y=nowy;
       c->icon_w=100;
       c->icontxt_x=nowt;
       XMapWindow(dpy,c->window);
       if(c->has_frame==0) {
         XUnmapWindow(dpy,c->top_left);
         XUnmapWindow(dpy,c->top_right);
         XUnmapWindow(dpy,c->title);
       }
    } else {
       nowx=c->x;
       nowy=c->y;
       noww=c->width;
       nowt=c->name_x;
       if( newc==c ) {
         c->is_icon=2;
       } else {
         c->is_icon=1; }
       if(c->has_frame==0) {
         XMapWindow(dpy,c->top_left);
         XMapWindow(dpy,c->top_right);
         XMapWindow(dpy,c->title);
       }
       newc=NULL;
       if(c->icon_x==-1000) {
         c->icon_x=nexticon_x;
         c->icon_y=nexticon_y;
         c->icon_w=100;
         c->icon_h=config.title.height;
         c->icontxt_x=3;
         c->icontxt_y=c->name_y;
         c->icontxt_w=c->name_w;
         c->icontxt_h=c->name_h;
         nexticon_y+=config.title.height;
       }
       endx=c->icon_x;
       endy=c->icon_y;
       endw=c->icon_w;
       endt=c->icontxt_x;
       c->ignore_unmap++;
       XUnmapWindow(dpy,c->window);
       for(i=0;i<ICONSTEPS;i++) {
         wc.width=noww-(noww-endw)*i/ICONSTEPS;
         wc.height=c->icon_h;
         wc.x=nowx-(nowx-endx)*i/ICONSTEPS;
         wc.y=nowy-(nowy-endy)*i/ICONSTEPS;
         c->icontxt_x=nowt-(nowt-endt)*i/ICONSTEPS;
         configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, True); 
         /*usleep(3000000);*/
       }
       wc.width=endw;
       wc.height=c->icon_h;
       wc.x=endx;
       wc.y=endy;
       c->icontxt_x=endt;
       c->is_icon=1;
   }

    shaded=get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED;

    if( c->is_icon==0 ) {
       raise_client(c);
       set_focus_client(c);
    }

    configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, ((shaded==WIN_STATE_SHADED) || (c->is_icon!=0))); 

    redraw_client_frame(c);

    if(c->is_icon!=0) {
      lower_client(c);

      newc=get_top_most_client(WIN_LAYER_NORMAL,False);
      if( newc==NULL )
        set_focus_client(NULL);
      else if( newc->is_icon!=0 )
        set_focus_client(NULL);
      else
        set_focus_client(newc);

      if(get_pointer_win(root)==c->frame) {
        c->is_icon=2;
        if( c->icontxt_w>=TW-6 )
          resize_client(c,c->icontxt_w+6-LW-RW+TLW+TRW,c->icon_h,True);
        redraw_client_frame(c);
        raise_client(c);
        overicon=c;
      } else
        c->is_icon=1;

    } else {
      raise_client(c);
      set_focus_client(c);
      if( shaded==WIN_STATE_SHADED )
        toggle_client_shaded(c);
    }
} 
 
#ifdef LAYER_SUPPORT 
/* 
 * 
 * Set client layer 
 * 
 */ 

void set_client_layer(Client *c, int layer) 
{ 
#ifdef DEBUG 
  printf("set_client_layer\n"); 
#endif 
 
  set_gnome_hint(c->window, WIN_LAYER, layer); 
  if(get_wm_state(c->window)==NormalState) 
    if(c->is_icon==0)
      raise_client(c); 
} 
 
/* 
 * 
 * Get client layer 
 * 
 */ 

int get_client_layer(Client *c) 
{ 
#ifdef DEBUG 
  printf("get client layer\n"); 
#endif 
 
  return get_gnome_hint(c->window, WIN_LAYER); 
} 
#endif /* LAYER */ 
 
/* 
 * 
 * Delete client 
 * 
 */ 

void delete_client(Client *c, int remap) 
{ 
#ifdef DEBUG 
  printf("delete_client (%#lx)\n",lastfocus); 
  printf("  deleting client [%#lx] %#lx\n",c->frame,c->window); 
#endif 
 
  /* Check if frame is still the clients parent */ 
  if(get_window_parent(c->window)==c->frame) 
  { 
    /* Reparent our client back to root if it still exists */ 
    gravitate_client(c, REMOVE_GRAVITY); 
    XReparentWindow(dpy, c->window, root, c->x, c->y); 
    set_wm_state(c->window, WithdrawnState); 

#ifdef KEYBOARD_SUPPORT
    /* Remove any key grabs on this window */
    ungrab_keys(c->window);
#endif
 
    /* Remap is true if we are quitting Oroborus */ 
    if(remap) 
      XMapWindow(dpy, c->window); 
    else 
      XUnmapWindow(dpy, c->window); 
  } 

#ifdef DEBUG
  printf("  checking lastenter: [%#lx], [%#lx]\n",c->frame,lastenter);
#endif
  if( c->frame==lastenter ) {
#ifdef DEBUG
    printf("  window was last to be entered: [%#lx]\n",lastenter);
#endif
    lastenter=0;
  }

  if(c->icon_y==nexticon_y-config.title.height)
    nexticon_y-=config.title.height;
	 
  /* Get rid of our frame window */ 
  XDestroyWindow(dpy, c->frame); 
 
  /* Remove reference from our client list */ 
  if(client_count==1) 
    client_list=NULL; 
  else 
  { 
    c->next->prev = c->prev; 
    c->prev->next = c->next; 
    if(c==client_list) 
      client_list=client_list->next; 
  } 
 
  if(c->name) XFree(c->name); 
  if(c->size) XFree(c->size); 
  free(c); 
 
  client_count--; 
  update_gnome_client_list();

#ifdef DEBUG
  printf("  client removed\n");
#endif
} 
 
/* 
 * 
 * Send close message to client window 
 * 
 */ 

void close_client(Client *c) 
{ 
  int i, n, found=False; 
  Atom *protocols=NULL; 
  XEvent e; 
 
#ifdef DEBUG 
  printf("close_client\n"); 
#endif 
 
  if(XGetWMProtocols(dpy, c->window, &protocols, &n)) 
  { 
    for(i=0; i<n; i++) 
      if(protocols[i]==icccm[WM_DELETE]) found=True; 
    if(protocols) XFree(protocols);  
  } 
  if(found) 
  { 
#ifdef DEBUG
    printf("  sending delete event to client [%#lx] %#lx\n",c->frame,c->window);
#endif
    e.type=ClientMessage; 
    e.xclient.window=c->window; 
    e.xclient.message_type=icccm[WM_PROTOCOLS]; 
    e.xclient.format=32; 
    e.xclient.data.l[0]=icccm[WM_DELETE]; 
    e.xclient.data.l[1]=CurrentTime; 
    XSendEvent(dpy, c->window, False, NoEventMask, &e); 
  } 
  else  {
#ifdef DEBUG
    printf("  killing client [%#lx] %#lx\n",c->frame,c->window);
#endif
    XKillClient(dpy, c->window); 
  }
} 
 
/* 
 * 
 * Redraw client borders 
 * 
 */ 

void redraw_client_frame(Client *c) 
{ 
  int state=FRAME_INACTIVE, shaded=0; 
 
#ifdef DEBUG 
  printf("redraw_client_frame [%#lx] %#lx (%#lx)\n",c->frame,c->window,lastfocus); 
#endif 
 
  if(!c->has_frame) 
  { 
#ifdef DEBUG 
    printf("  client (%#lx) has no frame\n", c->window); 
#endif 
    return; 
  } 

  if(c==get_focus_client()) 
    state=FRAME_ACTIVE;

/*  if(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED)*/
  if( c->is_icon==2 ) {
    shaded=2;
    state=FRAME_ACTIVE;
  } else if( c->is_icon==1 ) {
    shaded=2;
    state=FRAME_INACTIVE;
  }
 
  XSetWindowBackgroundPixmap(dpy, c->top_left, config.top_left.pixmap[state+shaded]); 
  XClearWindow(dpy, c->top_left); 
  XSetWindowBackgroundPixmap(dpy, c->top_right, config.top_right.pixmap[state+shaded]); 
  XClearWindow(dpy, c->top_right); 

  if( c->is_icon==0 ) {

    XSetWindowBackgroundPixmap(dpy, c->left, config.left.pixmap[state]); 
    XClearWindow(dpy, c->left); 
    XSetWindowBackgroundPixmap(dpy, c->right, config.right.pixmap[state]); 
    XClearWindow(dpy, c->right); 
    XSetWindowBackgroundPixmap(dpy, c->bottom_left, config.bottom_left.pixmap[state]); 
    XClearWindow(dpy, c->bottom_left); 
    XSetWindowBackgroundPixmap(dpy, c->bottom, config.bottom.pixmap[state]); 
    XClearWindow(dpy, c->bottom); 
    XSetWindowBackgroundPixmap(dpy, c->resize, config.resize.pixmap[state]); 
    XClearWindow(dpy, c->resize); 
 
    redraw_client_buttons(c); 
  }

  redraw_client_title(c); 
} 
 
/* 
 *  
 * Redraw client buttons 
 * 
 */ 

void redraw_client_buttons(Client *c) 
{ 
  int state=FRAME_INACTIVE; 
 
#ifdef DEBUG 
  printf("redraw_client_buttons [%#lx] %#lx (%#lx)\n",c->frame,c->window,lastfocus); 
  if(c->is_icon!=0)
    printf("  client is an icon - no drawing done\n");
#endif 
 
  if( c->is_icon!=0 )
    return;

  if(!c->has_frame) 
  { 
#ifdef DEBUG 
    printf("  client (%#lx) has no frame\n", c->window); 
#endif 
    return; 
  } 
 
  if(c==get_focus_client()) 
    state=FRAME_ACTIVE; 
 
  XSetWindowBackgroundPixmap(dpy, c->icon, config.icon.pixmap[c->icon_state==BUTTON_UNPRESSED?state:c->icon_state]); 
  XClearWindow(dpy, c->icon); 
  XSetWindowBackgroundPixmap(dpy, c->close, config.close.pixmap[c->close_state==BUTTON_UNPRESSED?state:c->close_state]); 
  XClearWindow(dpy, c->close); 
  if(c->can_max) {
    if(c->hide_state!=BUTTON_UNPRESSED) state=c->hide_state;
  } else if(state==FRAME_ACTIVE) {
    state=BUTTON_ACTIVEOFF;
  }
  XSetWindowBackgroundPixmap(dpy, c->hide, config.hide.pixmap[state]);
  XClearWindow(dpy, c->hide);

} 
 
/* 
 * 
 * Redraw client name in the title bar 
 * 
 */ 

void redraw_client_title(Client *c) 
{ 
  int state=FRAME_INACTIVE, shaded=0; 
  Pixmap temp; 
  GC gc; 
  XGCValues gv; 
#ifdef I18N_SUPPORT 
  XRectangle ink, logical; 
  int ascent=0, descent=0; 
#endif /* I18N */ 
  int width; 
 
#ifdef DEBUG 
  printf("redraw_client_title [%#lx] %#lx (%#lx)\n",c->frame,c->window,lastfocus); 
#endif 
 
  if( !c->has_frame && (c->is_icon==0) ) 
  { 
#ifdef DEBUG 
    printf("  client (%#lx) has no frame\n", c->window); 
#endif 
    return; 
  }
 
  if(c==get_focus_client()) 
    state=FRAME_ACTIVE; 

/*  if(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED)*/
  if( c->is_icon==1 ) {
    shaded=2;
    state=FRAME_INACTIVE;
  } else if( c->is_icon==2 ) {
    shaded=2;
    state=FRAME_ACTIVE;
  }
 
  gv.function=GXcopy; 
  gv.tile=config.title.pixmap[state+shaded]; 
  gv.fill_style=FillTiled; 
  gc=XCreateGC(dpy, root, GCFunction|GCFillStyle|GCTile, &gv); 
  if( c->is_icon!=0 ) {
    temp=XCreatePixmap(dpy, root, TW, TH, DefaultDepth(dpy, screen));
    XFillRectangle(dpy, temp, gc, 0, 0, TW, TH); 
  } else {
    temp=XCreatePixmap(dpy, root, TW, TH, DefaultDepth(dpy, screen)); 
    XFillRectangle(dpy, temp, gc, 0, 0, TW, TH); 
  }
 
  if(c->name) 
  { 
    int posx=0;
    int posy=0; 
#ifdef I18N_SUPPORT
    if(use_fontset)
    {
      XmbTextExtents(config.fontset, c->name, strlen(c->name), &ink, &logical); 
      ascent=-logical.y; 
      descent=logical.height+logical.y; 
      width=logical.width; 
    }
    else
      width=XTextWidth(config.font, c->name, strlen(c->name));
#else  
    width=XTextWidth(config.font, c->name, strlen(c->name)); 
#endif /* I18N */ 
 
    switch(config.text_align) 
    { 
    case ALIGNMENT_MIDDLE: 
      posx=width>TW-2?2:TW/2-width/2; 
      break; 
    case ALIGNMENT_RIGHT: 
      posx=width>TW-2?2:TW-width; 
      break; 
    case ALIGNMENT_LEFT: 
      posx=13; 
      break; 
    } 
    if( c->is_icon!=0 )
       posx=3;
    else if( posx<config.close.width+config.hide.width+config.icon.width-1 )
       posx=config.close.width+config.hide.width+config.icon.width-1;

#ifdef I18N_SUPPORT
    if(use_fontset)
      posy=(TH-descent)-((TH-ascent-descent)/2);
    else
      posy=(TH-config.font->descent)-((TH-config.font->ascent-config.font->descent)/2);
#else 
    posy=(TH-config.font->descent)-((TH-config.font->ascent-config.font->descent)/2); 
#endif /* I18N */ 

    if( c->is_icon!=0 ) {
      if( c->icontxt_w==width ) {
         posx=c->icontxt_x;
         posy=c->icontxt_y;
      } else {
         c->icontxt_x=posx;
         c->icontxt_y=posy;
         c->icontxt_w=width; }
    } else if( c->name_w==width && c->name_x==posx && c->name_y==posy ) {
         posx=c->name_x;
         posy=c->name_y;
    } else {
         c->name_x=posx;
         c->name_y=posy;
         c->name_w=width;
    }

#ifdef I18N_SUPPORT
    if(use_fontset)
      XmbDrawString(dpy, temp, config.fontset, config.font_gc[state], posx, posy, c->name, strlen(c->name));
    else
      XDrawString(dpy, temp, config.font_gc[state], posx, posy, c->name, strlen(c->name));
#else 
    XDrawString(dpy, temp, config.font_gc[state], posx, posy, c->name, strlen(c->name)); 
#endif /* I18N */ 
 
  }
  XSetWindowBackgroundPixmap(dpy, c->title, temp); 
  XClearWindow(dpy, c->title); 
 
  XFreeGC(dpy, gc); 
  XFreePixmap(dpy, temp); 
} 
 
#ifdef SHAPE_SUPPORT 
/* 
 * 
 * Set client window shape 
 * 
 */ 

void set_client_shape(Client *c) 
{ 
  Window temp; 
 
#ifdef DEBUG 
  printf("set_client_shape [%#lx] %#lx (%#lx)\n",c->frame,c->window,lastfocus); 
#endif 
 
  if(shape) 
  { 
    temp=XCreateSimpleWindow(dpy, root, 0, 0, FW, FH, 0, 0, 0);
    XShapeCombineShape(dpy, temp, ShapeBounding, LW, TH, c->window, ShapeBounding, ShapeSet);

    if(c->has_frame || (c->is_icon!=0)) 
    {
/*      if(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED)*/
      if( c->is_icon!=0 ) {
        XShapeCombineMask(dpy, c->top_left, ShapeBounding, 0, 0, config.top_left.mask[1], ShapeSet);
        XShapeCombineMask(dpy, c->top_right, ShapeBounding, 0, 0, config.top_right.mask[1], ShapeSet);
        XShapeCombineMask(dpy, temp, ShapeBounding, TLX, TLY, config.top_left.mask[1], ShapeUnion);
        XShapeCombineMask(dpy, temp, ShapeBounding, TRX, TRY, config.top_right.mask[1], ShapeUnion);
        XShapeCombineShape(dpy, temp, ShapeBounding, TX, TY, c->title, ShapeBounding, ShapeUnion);  
        XShapeCombineShape(dpy, temp, ShapeBounding, LX, LY, c->left, ShapeBounding, ShapeUnion); 
        XShapeCombineShape(dpy, temp, ShapeBounding, RX, RY, c->right, ShapeBounding, ShapeUnion); 
        XShapeCombineShape(dpy, temp, ShapeBounding, BX, BY, c->bottom, ShapeBounding, ShapeUnion);
      } else {
        XShapeCombineMask(dpy, c->bottom_left, ShapeBounding, 0, 0, config.bottom_left.mask[0], ShapeSet);
        XShapeCombineMask(dpy, c->resize, ShapeBounding, 0, 0, config.resize.mask[0], ShapeSet);
        XShapeCombineMask(dpy, c->hide, ShapeBounding, 0, 0, config.hide.mask[0], ShapeSet);
        XShapeCombineMask(dpy, c->icon, ShapeBounding, 0, 0, config.icon.mask[0], ShapeSet);
        XShapeCombineMask(dpy, c->close, ShapeBounding, 0, 0, config.close.mask[0], ShapeSet);
        XShapeCombineMask(dpy, temp, ShapeBounding, TLX, TLY, config.top_left.mask[0], ShapeUnion);
        XShapeCombineMask(dpy, temp, ShapeBounding, TRX, TRY, config.top_right.mask[0], ShapeUnion);
        XShapeCombineMask(dpy, temp, ShapeBounding, RSX, RSY, config.resize.mask[0], ShapeUnion); 
        XShapeCombineMask(dpy, temp, ShapeBounding, BLX, BLY, config.bottom_left.mask[0], ShapeUnion);
        XShapeCombineShape(dpy, temp, ShapeBounding, TX, TY, c->title, ShapeBounding, ShapeUnion);  
        XShapeCombineShape(dpy, temp, ShapeBounding, LX, LY, c->left, ShapeBounding, ShapeUnion); 
        XShapeCombineShape(dpy, temp, ShapeBounding, RX, RY, c->right, ShapeBounding, ShapeUnion); 
        XShapeCombineShape(dpy, temp, ShapeBounding, BX, BY, c->bottom, ShapeBounding, ShapeUnion);
      }
    } 
    XShapeCombineShape(dpy, c->frame, ShapeBounding, 0, 0, temp, ShapeBounding, ShapeSet); 
    XDestroyWindow(dpy, temp);
  } 
} 
#endif /* SHAPE */ 
 
#ifdef WORKSPACE_SUPPORT 
/* 
 * 
 * Toggle sticky/unsticky 
 * 
 */ 

void toggle_client_sticky(Client *c) 
{ 
#ifdef DEBUG 
  printf("toggle_client_sticky [%#lx] %#lx (%#lx)\n",c->frame,c->window,lastfocus); 
#endif 
 
  set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_STICKY); 
} 
 
/* 
 * 
 * Move client to new workspace 
 * 
 */ 

void move_client_to_workspace(Client *c, int ws) 
{ 
#ifdef DEBUG 
  printf("move_client_to_workspace\n"); 
#endif 
 
  if(ws==workspace || ws<0 || ws>=config.workspace_count) return; 
  set_gnome_hint(c->window, WIN_WORKSPACE, ws); 
  change_to_workspace(ws); 
  if(c->is_icon==0) {
    set_focus_client(c); 
    raise_client(c); }
} 
#endif /* WORKSPACE */ 

